from datetime import datetime
from typing import Any, Optional, Tuple, Union


ALLOWED_KINDS = (
    "gsv",
    "opa",
    "par",
)

class RequestIdentifierError(ValueError):
    def __init__(self, *args):
        self.identifier = args[0]
        msg = (
            f"Malformed request: {self.identifier}. "
            "It should be in one of the following forms: "
            "{'MM/REQID', 'YYMM/REQID', 'REQID'}."
        )
        super().__init__(msg, *args[1:])


def _normalize_request_name(name: Any, subrequest: Optional[Any] = None) -> Tuple[str, Union[str, None]]:
    if isinstance(name, float):
        raise NotImplementedError(
            "Passing request name as float not implemented yet. "
            "Please, pass as a string with a dot."
        )
    if subrequest is not None:
        subrequest = str(subrequest)
    if isinstance(name, int):
        return str(name), subrequest
    if isinstance(name, str):
        if "." in name:
            name, subrequest = name.split(".", 1)
        return name, subrequest
    raise TypeError(
        "Request/subrequest names must be expressed as "
        f"int, float or str, but were {type(name)}/{type(subrequest)}."
    )

def _validate_month(month: str):
    if not 1 <= int(month) <= 12:
        raise ValueError(f"Bad value for month: {month}.")

def _normalize_request_month(month: str | int | None, identifier: Any) -> str:
    now = datetime.now()
    if month is None:
        return now.strftime("%y%m")
    month = str(month)
    if month == "*":
        return month
    if not month.isdigit():
        raise ValueError(f"Value for month must be a number, but was '{month}'.")
    if len(month) == 2:
        _validate_month(month)
        return now.strftime("%y") + month
    elif len(month) == 4:
        _validate_month(month[2:4])
        return month
    raise RequestIdentifierError(identifier)

def normalize_request_identifier(
    identifier: Any,
    /,
    subrequest: Optional[Any] = None,
    month: Optional[Union[str, int]] = None,
    kind: str = "*"
) -> Tuple[str, Union[str, None], str, str]:
    # Split identifier
    if isinstance(identifier, str) and "/" in identifier:
        month, name = identifier.split("/", 1)
    else:
        name = identifier
        month = _normalize_request_month(month, identifier)
    # Normalize triplet
    name, subrequest = _normalize_request_name(name, subrequest)
    month = _normalize_request_month(month, identifier)
    if kind not in ["*", *ALLOWED_KINDS]:
        raise ValueError(f"Wrong kind: '{kind}', must be in {ALLOWED_KINDS}.")
    return name, subrequest, month, kind

